package project;

import java.awt.*;
import java.lang.String;
import java.io.*;
import java.lang.Math;
import java.lang.InterruptedException;

/**
 *	Class used in animating the Dining Philosophers Algorithm visually.
 *	One of these is shared between all the Philosophers used in one
 *	algorithm. They will call the animate function of this class seperatly
 *	specifing their identity through an integer-parameter. Using this info
 *	the class will repaint only the calling philosopher without interfering
 *	with other philosophers
 *
 *	@author Filip Van Rysselberghe
 */
public class AlgDiningPhilAnimDrawer extends AlgAnimDrawer {

	/**
	 *	Constructor. Will invoke his super (AlgAnimDrawer) using the
	 *	provided parameters.
	 *
	 *	@param AlgThread		alg thread that's drawing
	 *	@param AlgAnimatorConfig	configuration for the animated algorithm
	 */
	public AlgDiningPhilAnimDrawer(AlgThread thread, AlgAnimatorConfig config){
		//invoke constructor on super
		super(thread, config);
	}

	/**
	 *	Initialises the table with 5 plates of thinking philosophers
	 *	So one table will be drawn and 5 plates on it.
	 */
	public void animate(){
		Image 		image = fThread.fImages.getAnimationImage();
		Graphics 	g     = image.getGraphics();

		//calculate which one of the height or length has the smallest
		//value. We'll use this value to calculate the diameter of the
		//table.
		int 	border = 10,
				width  = ((fConfig.width/2) - 4) - (2*border),
               	height = (fConfig.height-90) - (2*border);

		int diameter = (width < height) ? width : height;

		//draw the table
		drawTable(g, diameter/2);

		//initialize table with 5 plates
		drawPlate(g, 0, diameter/2, 0);
		drawPlate(g, 1, diameter/2, 0);
		drawPlate(g, 2, diameter/2, 0);
		drawPlate(g, 3, diameter/2, 0);
		drawPlate(g, 4, diameter/2, 0);

		fThread.fImages.paintAnimation(image);
	}

	/**
	 *	Animation function that will be called by a Philosopher. Using
	 *	the provided plateNr (actually the identity of the philosopher)
	 *	the right plate will be redraw in the specified actionColor.
	 *	Only the specified plate will be redrawn, all other plates
	 *	will be left untouched
	 *
	 *	@param int		number of the plate (id of philosopher)
	 *	@param int		action the philosopher is performing
	 */
	public void animate(int plateNr, int action) throws InterruptedException {
		Image 		image = fThread.fImages.getAnimationImage();
		Graphics 	g     = image.getGraphics();

		//calculate which one of the height or length has the smallest
		//value. We'll use this value to calculate the diameter of the
		//table.
		int 	border = 10,
				width  = ((fConfig.width/2) - 4) - (2*border),
                height = (fConfig.height-90) - (2*border);

		int diameter = (width < height) ? width : height;

		//redraw plate with right actioncolor
		drawPlate(g, plateNr, diameter/2, action);

		fThread.fImages.paintAnimation(image);

		this.sleep();
	}

	/**
	 *	Function that draws the table on the screen
	 *
	 *	@param Graphics			graphics on which table will be drawn
	 *	@param radius			radius of the table we'de like
	 */
	private void drawTable(Graphics g, int radius){
		//calculate coordinates of the startingpoint of our table
		int	border = 10,
			x = border,
			y = border,
			width  = 2 * radius,
			height = width;

		//using this info we can a circle representing our table
		g.setColor( Color.black );
		g.drawOval(x, y, width, height);
	}

	/**
	 *	Draws the specified plate in the according actioncolor.
	 *
	 *	@param Graphics			Graphics on which we 'll draw
	 *	@param int			number of the plate
	 *	@param radiusTable		radius of the table on which we place
	 *					the plate. The size of the plate will
	 *					calculated using that info.
	 *	@param int			action of the plate
	 */
	private void drawPlate(Graphics g, int plateNr, int radiusTable, int action){
		//calculate middle of table
		int 	border = 10,
			xMiddleTable = border + radiusTable,
			yMiddleTable = xMiddleTable;

		//we use these coordinates to calculate the middle
		//of the plate.
		Point 	middle = new Point(calculateMiddle(plateNr, xMiddleTable, yMiddleTable, ((3*radiusTable)/5) ));

		//these values can be used to calculate the real x, y
		//coordinates we need for drawing the circle
		int	radiusPlate = (radiusTable / 5);

		//when drawing we need to position in the right top corner
		//so we translate our middle to that point
		middle.translate(-radiusPlate, -radiusPlate);
		int	width  = 2*radiusPlate,
			height = width;

		//using this info we can a circle representing our table and
		//philospoher. Depending on the current action of our philospher, the
		//plate will get another color
		if ( action == 0 ) {			//thinking
			g.setColor( Color.red );
		}
		else if ( action == 1 ) {		//waiting for forks
			g.setColor( Color.orange );
		}
		else {					//eating
			g.setColor( Color.green );
		}

		g.fillOval(middle.x, middle.y, width, height);

	}

	/**
	 *	Calculates the coordinates of the middle of the plate with specified number
	 *
	 *	@param int			number of plate
	 *	@param int			x-coordinate of the middle of the table
	 *	@param int			y-coordinate of the middle of the table
	 *	@param int			radius of table
	 *
	 *	@return Point			coordinates of the middle of the plate
	 */
	private Point calculateMiddle(int number, int xMiddle, int yMiddle, int radius){
		//opgepast dit is onveilige methode

		//zoek het snijpunt van een cirkel met zelfde middelpunt als
		//tafel maar met als straal 3/5 van de straal van de tafel,
		//met de rechte door het middelpunt vande tafel naar een hoek
		//punt van een regelmatige vijfhoek.
		double 	angle = ((2*Math.PI)/5)*number,
			rc = ( Math.sin(angle) )/ ( Math.cos(angle) );

		//vergelijking van de rechte = (x-xMiddle)rc = (y- yMiddle)


		//vergelijking van de cirkel = (x-xMiddle)2 + (y-yMiddle)2 = radius2

		//op te lossen vgl naar x = (1+ rc2)x2 - 2*(1+rc2)xxMiddle + (1+rc2)xMiddle2 = radius2
		double	a =  (1 + Math.pow(rc,2)),
			b = (-2) * ((1 + Math.pow(rc,2)) * xMiddle),
			c =  ( (1 + Math.pow(rc,2)) * Math.pow(xMiddle, 2) ) - Math.pow(radius, 2);

		double 	desc = Math.pow(b,2)-4*a*c;

		Point middle = new Point();

		//choose wich null-point we have to take
		if ( Math.cos(angle) > 0 ) {
			middle.x = (int) ( ( -b + Math.sqrt(desc) ) / ( 2*a ) );
		}
		else {
			middle.x = (int) ( ( -b - Math.sqrt(desc) ) / ( 2*a ) );
		}
		middle.y = (int) ( ( middle.x - xMiddle )*rc + yMiddle );

		return middle;
	}
}
