package project;

import java.lang.Thread;
import java.lang.InterruptedException;
import java.lang.Math;

/**
 *	Philosopher as used in the Dining Philospher algorithm. We solve
 *	this IPC-problem using Semaphores. 
 *	States: Thinking = 0, Hungry = 1 and Eating = 2;
 *
 *	@author Filip Van Rysselberghe
 */

class Philosopher extends Thread{
	//who am I
	private int fIdentity;

	//reference to needed semaphores and states.
	private int[] rState = new int[5];
	private Semaphore rMutex;
	private Semaphore[] rS = new Semaphore[5];

	//drawer for animation
	private AlgDiningPhilAnimDrawer fAnimDraw;
	
	/*
	 *	Constructor which will fill the content of the Philosopher object
	 *	with an Id and references to the shared information between the
	 *	philosphers such as the mutex, state of each philosopher
	 *
	 *	@param int		id of philosopher
	 *	@param int[]		refernce to the state of all philosophers
	 *	@param Semaphore	reference to the mutex
	 *	@param Semaphore[]	reference to the fork-semaphore
	 *	@param AlgDiningPhilAnimDrawer	refernce to drawer used when animating 
	 */
	public Philosopher(int id, int[] state, Semaphore mutex, Semaphore[] s, AlgDiningPhilAnimDrawer animDraw){
		fIdentity = id;

		rState = state;
		rMutex = mutex;
		rS = s;
		
		fAnimDraw = animDraw;
	}
	
	/**
	 *	The life of our Philosopher. Executed when the start-method of
	 *	the object is called. A philosopher thinks and eats.  
	 */
	public void run() {
		try { 
			while (true) {
				think();
				take_forks(fIdentity);
				eat();
				put_forks(fIdentity);
			}
		}
		catch ( InterruptedException e) {
		}
	}

	/**
	 *	The favourite action of our Philosopher.
	 */
	private void think() throws InterruptedException{
		//philosopher has no forks and is thinking
		fAnimDraw.animate(fIdentity, 0);
	}

	/**
	 *	The necessary action to stay alive.
	 */
	private void eat() throws InterruptedException{
		//philosopher has acquired forks and is eating his delicous 
		//spaghetti
		fAnimDraw.animate(fIdentity, 2);
	}

	/**
	 *	Before one can eat, he must acquire two forks
	 *
	 *	@param int		one who wants to eat
	 */
	private void take_forks(int i) throws InterruptedException{
		rMutex.down();
			rState[i] = 1;
			test(i);
		rMutex.up();
		
		//philosopher is waiting for his forks and is fucking hungry
		fAnimDraw.animate(fIdentity, 1);
		
		rS[i].down();
		
	}

	/**
	 *	After eating a philosopher doesn't do any dishes. He just puts 
	 *	down his forks. Neigbours can take a fork if the fork was shared
	 *
	 *	@param int		philosopher putting down his fork
	 */
	private void put_forks(int i) throws InterruptedException{
		rMutex.down();
			rState[i] = 0;
			//Philosopher will release forks and start thinking
			//the animate happens before real release of forks
			//because the left/right neighbour could be waiting
			//for his forks and will immediatly turn eating after 
			//release. This would give an inconsistent animation. 
			//(neighbouring plates with same eating color)
			fAnimDraw.animate(fIdentity, 0);
			
			test( ( Math.abs(i-1) ) %5 );
			
			test( ( Math.abs(i+1) ) %5 );
		rMutex.up();
	}

	/**
	 *	Test wether he can take the forks by looking at his neighbours and
	 *	chenking wether they are eating
	 *
	 *	@param int		philosopher checking out his neighbours
	 */
	private void test(int i){
		if ( (rState[i] == 1) && (rState[ ( Math.abs(i-1) )%5 ] != 2) && (rState[ ( Math.abs(i+1) )%5 ] != 2) ){
			rState[i] = 2;
			rS[i].up();
		}
	}

}
