package project;

import java.lang.InterruptedException;

/**
 *	This class represents the system of Semaphores as proposed by
 *	E.W. Dijkstra in 1965. He suggested using an integer variable to
 *	count the number of wake-ups saved for future use. He also proposed
 *	2 functions with the original names P( proberen ) and V (verhogen).
 *	Since these two Dutch words were translated into the more international
 *	Down and Up, we provided these english variants and not the Dutch ones.
 *
 *	@author Filip Van Rysselberghe
 */
public class Semaphore {

	/**
	 *	number of wake-ups saved for future use
	 */
	private int fValue = 0;

	/**
	 *	Default constructor. No wake-ups were saved. So in case
	 *	of a down this Semaphore will block until an up is executed
	 */
	public Semaphore(){
	}

	/**
	 *	Using this second constructor you can specify the number of
	 *	wake-ups that were previously saved. This can be handy when
	 *	using a semaphore as a Mutex. By creating a Semaphore with
	 *	value 1, you can simulate a Mutex. The Down function can be
	 *	called when entering the critical region. Doing so will block
	 *	all other threads wanting to enter the critical region
	 *	The up will release the lock and will be called when leaving the
	 *	critical section
	 *
	 *	@param int		number of wake-up the semaphore has saved
	 */
	public Semaphore(int value) {
		fValue = value;
	}


	/**
	 *	The Up function is the Semaphore equivalent of the ordinary
	 *	wakeup. It will increment the number of Wake-ups stored and
	 *	will notify blocked Threads that one of them can continue.
	 *
	 *	@see down()
	 */
	public synchronized void up() {

		++fValue;
		//notify interested threads
		notify();

	}

	/**
	 *	It's the semaphore equivalent of the Sleep-function. When called
	 *	It 'll check wether the semaphore value is bigger than 0. If so,
	 *	it decrements the value(uses up one saved wake-up)and just continues
	 *	execution. In the other case the process will be put to sleep (wait)
	 *	and will be reactivated upon an up().
	 *	Checking the value, changing it and possibly going to sleep happens
	 *	in one atomic action.
	 *
	 *	@see up()
	 */
	public synchronized void down() {
		//check value bigger then 0
		//if so decrement value and continue
		if ( fValue > 0 ) {
			//just continue
		}
		else {
			//sleep and possibility to continue after an "up"
			try {
				wait();
			}
			catch ( InterruptedException e) {
			}
		}
		//the original design of Dijkstra only went till a value 0
		//in the case of zero the thread accessing was put asleep and
		//after an up it was waken up and then the value would be
		//decreased. This is the same as always decreasing and checking
		//for negative numbers
		--fValue;
	}

}
